/******************************************************************/
/*									                              */
/*      Description : MDS process                                 */
/*      Making Date : 2010. 05. 06					              */
/*      Affiliate   : Machine Learning Lab in SoongSil Univ       */
/*      Developer   : Sungin Park(rfchief@gmail.com)              */
/******************************************************************/

#include <stdio.h>
#include <cublas.h>
#include <cuda.h>
#include <culapack.h>
#include "common.h"
#include "cMDS.h"
#include "utilFunc.h"

__global__ void squareM(float* A, float* B, int width, int height)
{   
   unsigned int xIndex = blockDim.x * blockIdx.x + threadIdx.x;
   unsigned int yIndex = blockDim.y * blockIdx.y + threadIdx.y;
   
   if ((xIndex < width) && (yIndex < height))
   {
       unsigned int index_in  = xIndex + width * yIndex;       
       B[index_in] = (A[index_in] * A[index_in]);       
   }	
}

//Classical MDS
float* cMDS (int pSize_x, int pSize_y, int dim, float *inputData, struct paramsStr params)
{	
	///////////////////////////////////////////////////////
	//////Initialize part
	////////////////////////////////////////////////////////
    int size_x = pSize_x;                                             //row
	int size_y = pSize_y;                                             //col
	int dimension = dim;                                              //reduction dimension
	int t_size = size_x * size_y;
	int memSize = t_size * sizeof(float);
	
	int i = 0, j = 0;
    int tempModX = 0, tempModY = 0;    
    int bSize_x = 0;
    int bSize_y = 0;
	char jobu = 'A';
    char jobvt = 'A';
	
	float *h_PMatrix    = inputData;                                           //Proximity Matrix
	float *h_jMatrix    = (float *)malloc(memSize);                            //J Matrix                           J = I - (1/n)*1*1'
	float *h_DCMatrix   = (float *)malloc(memSize);                            //Double Centering Matrix            B = -1/2Jp^2J
	float *h_coordinate = (float *)malloc(size_x*dimension*sizeof(float));     //MDS coordinate
    
    float* S = NULL;
    float* U = NULL;
    float* VT = NULL;
	float* eValue = NULL;
	float* eVector = NULL;

    S       = (float*)malloc(imin(size_x,size_y)*sizeof(float));                //eigenvalues of SVD
    eValue  = (float*)malloc(dimension*dimension*sizeof(float));                
    U       = (float*)malloc(size_x*size_x*sizeof(float));                      //eigenvectors of SVD
	eVector = (float*)malloc(size_x*dimension*sizeof(float));                   
    VT      = (float*)malloc(size_x*size_x*sizeof(float));

    float *d_data1; 
    float *d_data2;
    float *d_result;
	
	cublasStatus status;
	culaStatus cStatus;
    
////////////////////////////////////////////////////////////////////////  

    /* Initialize CUDA */
	//memory Allocate
	cudaMalloc((void**) &d_data1, memSize);
	cudaMalloc((void**) &d_result, memSize);    
	cudaMemcpy(d_data1, h_PMatrix, memSize, cudaMemcpyHostToDevice);	
	dim3 threads(BLOCK_SIZE, BLOCK_SIZE);
		
    tempModX = size_x % BLOCK_SIZE;
    tempModY = size_y % BLOCK_SIZE;

    if ((tempModX) == 0)
    {
        bSize_x = size_x / BLOCK_SIZE;
    }
    else
    {
        bSize_x = size_x / BLOCK_SIZE + 1;
    }

    if ((tempModY) == 0)
    {
        bSize_y = size_y / BLOCK_SIZE;
    }
    else
    {
        bSize_y = size_y / BLOCK_SIZE + 1;
    }

    dim3 grid(bSize_x, bSize_y);
    
	//Execute Kernel
    squareM<<<grid, threads >>>(d_data1, d_result, size_x, size_y);

    // copy result from device to host
    cudaMemcpy(h_PMatrix, d_result, memSize, cudaMemcpyDeviceToHost);

	/* Allocate device memory for the matrices */
    status = cublasAlloc(t_size, sizeof(float), (void**)&d_data1);
    cublasCheckStatus(status, params);
    status = cublasAlloc(t_size, sizeof(float), (void**)&d_data2);
    cublasCheckStatus(status, params);
    status = cublasAlloc(t_size, sizeof(float), (void**)&d_result);
    cublasCheckStatus(status, params);

	//Construct J Matrix	
    for (i = 0; i < t_size; i += (size_x+1))
    {
        h_jMatrix[i] = 1.0f - ((float)(1.0/size_x));
		for(j = i+1; j< i+(size_x+1); j++)
		{
			if(j < t_size){
                h_jMatrix[j] = 0.0f - ((float)(1.0/size_x));
			}
		}//for            
    }//for
       

	//////////////////////////////////////////////////////////////////////////////
	//Double Centering Process Start
	///////////////////////////////////////////////////////////////////////////
	status = cublasSetVector(t_size, sizeof(float), h_PMatrix, 1, d_data1, 1);
    cublasCheckStatus(status, params);
    status = cublasSetVector(t_size, sizeof(float), h_jMatrix, 1, d_data2, 1);
    cublasCheckStatus(status, params);
    status = cublasSetVector(t_size, sizeof(float), h_jMatrix, 1, d_result, 1);
    cublasCheckStatus(status, params);                 


    cublasSgemm('n', 'n', size_x, size_y, size_y, 1.0f, d_data2, size_x, d_data1, size_x, 0.0f, d_result, size_x);                 //J*P            data1 : P^2 Matrix, Data2 : jMatrix
	cublasSgemm('n', 'n', size_x, size_y, size_y, 1.0f, d_result, size_x, d_data2, size_x, 0.0f, d_data1, size_x);                 //(J*P)*J        data1 : J*P Matrix  Data2 : jMatrix
    cublasSscal(t_size, -0.5f, d_data1, 1);                                                                                        //-1/2 * (JPJ)
	cublasGetVector(t_size, sizeof(float), d_data1, 1, h_DCMatrix, 1);
    
	status = cublasFree(d_data1);
    cublasCheckStatus(status, params); 
	status = cublasFree(d_data2);
    cublasCheckStatus(status, params); 
	status = cublasFree(d_result);
    cublasCheckStatus(status, params); 

	///////////////////////////////////////////////////////
	//Double Centering Process End
	//////////////////////////////////////////////////////////

	///////////////////
	////SVD Process
	////////////////////
    
	//compute SVD in CULA
    cStatus = culaSgesvd(jobu, jobvt, size_x, size_y, h_DCMatrix, size_x, S, U, size_x, VT, size_y);
    checkStatus(cStatus, params);
    //SVD Process end

	///////////////////
	////SVD Process end
	////////////////////

	/////////////////////////////////////////////////
	//final part of MDS
	///////////////////////////////////////////////  
	//eigenvalue
	for(i = 0 ; i<dimension; i++)
	{
		for(j = 0;j < dimension;j++)
		{
			if(i==j){
				eValue[(i*dimension)+j] = sqrt(S[i]);
			}else{
				eValue[(i*dimension)+j] = 0.0f;
			}
		}
	}
	
	//eigenvector
	for(i=0; i<dimension*size_x; i++)
	{		
		eVector[i] = U[i];	
	}	

    status = cublasAlloc(dimension*size_x, sizeof(float), (void**)&d_data1);
    cublasCheckStatus(status, params); 
	status = cublasAlloc(dimension*dimension, sizeof(float), (void**)&d_data2);
    cublasCheckStatus(status, params); 
    status = cublasAlloc(dimension*size_x, sizeof(float), (void**)&d_result);
    cublasCheckStatus(status, params); 
	
    status = cublasSetVector(dimension*size_x, sizeof(float), eVector, 1, d_data1, 1);
    cublasCheckStatus(status, params); 
    status = cublasSetVector(dimension*dimension, sizeof(float), eValue, 1, d_data2, 1);
    cublasCheckStatus(status, params); 

    cublasSgemm('n', 'n', size_x, dimension, dimension, 1.0f, d_data1, size_x, d_data2, dimension, 0.0f, d_result, size_x);
    cublasGetVector(dimension*size_x, sizeof(float), d_result, 1, h_coordinate, 1);

	cublasFree(d_data1);
	cublasFree(d_data2);
    cublasFree(d_result);

	
	/////////////////////////////////////////////////
	//final part of MDS
	///////////////////////////////////////////////  
    
    /* Clean up workspace, input, and output memory allocations */	
	free(VT);
    free(U);
    free(S);

	free(h_jMatrix);
	free(h_DCMatrix);
	free(eValue);
	free(eVector);
    
	return h_coordinate;
}//cMDS
